<?php

/**
 * JCH Optimize - Joomla! plugin to aggregate and minify external resources for
 * optmized downloads
 * @author Samuel Marshall <sdmarshall73@gmail.com>
 * @copyright Copyright (c) 2010 Samuel Marshall
 * @license GNU/GPLv3, See LICENSE file
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * If LICENSE file missing, see <http://www.gnu.org/licenses/>.
 *
 * This plugin, inspired by CssJsCompress <http://www.joomlatags.org>, was
 * created in March 2010 and includes other copyrighted works. See individual
 * files for details.
 */
// no direct access
defined('_JEXEC') or die('Restricted access');

jimport('joomla.plugin.plugin');

if (!defined('DS'))
{
        define('DS', DIRECTORY_SEPARATOR);
}

class JCH_Optimize extends JPlugin
{

        /**
         * Search area used to find js and css files to remove
         * 
         * @param string $sHead         Head of document
         * @return type
         */
        protected function getSearchArea($sHead)
        {
                return $sHead;
        }

        /**
         * Fetches HTML to be sent to browser
         * 
         * @return string
         */
        protected function getBody()
        {
                return $sBody = preg_replace($this->getHeadRegex(), $this->sSearchArea, $this->sBody);
        }

        /**
         * Regex to find js links in search area
         * 
         * @return array
         */
        protected function getJsRegex()
        {
                $aJsRegex          = array();
                $aJsRegex['start'] = '
                        ~<script
                                (?=[^>]+src
                                        (?=[^>]*[=/"\']([^=/"\']+\.js))
                                        (?:=["\']?([^"\'\s>]+))
                                )
                                (?:(?!(?:type=(?!["\']?text/javascript))';
                $aJsRegex['end']   = ''
                        . '     )[^>])*'
                        . '/?></script>~isx';

                return $aJsRegex;
        }

        /**
         * Regex to find css links in search area to remove
         * 
         * @return array
         */
        protected function getCssRegex()
        {
                $aCssRegex          = array();
                $aCssRegex['start'] = '
                        ~(?:<link
                                (?=[^>]+href
                                        (?=[^>]*[=/"\']([^=/"\']+\.css))
                                        (?:=["\']?([^"\'\s>]+))
                                )
                                (?:(?!(?:itemprop)
                                        |(?:disabled)
                                        |(?:type=(?!["\']?text/css))
                                        |(?:rel=(?!["\']?stylesheet))';
                $aCssRegex['end']   = ''
                        . '     )[^>])*'
                        . '>)~ix';

                return $aCssRegex;
        }

        /**
         * 
         * @param type $sContent
         * @return type
         */
        protected function replaceImports($sContent)
        {
                return $sContent;
        }

        /**
         * Determines if file requires http protocol to get contents (Not allowed)
         * 
         * @param string $sUrl
         * @return boolean
         */
        protected function isUrlFopenAllowed($sUrl)
        {
                return ((preg_match('~(?:.*\.php)~', $sUrl) === 1) || !$this->isInternal($sUrl));
        }

        /**
         * 
         * @param type $sPath
         * @return boolean
         */
        protected function loadAsync($sPath)
        {
                return false;
        }

        /**
         * Excludes js files in editor folders
         * 
         * @param string $sUrl  File url
         * @return boolean
         */
        protected function isEditorsExcluded($sUrl)
        {
                return (preg_match('~.*/editors/~i', $sUrl));
        }

        /**
         * 
         * @return string
         */
        protected function getAsyncAttribute()
        {
                return '';
        }

}

class plgSystemJCH_Optimize extends JCH_Optimize
{

        /** @var object   Holds reference to JCACHE object */
        protected $oCache = '';

        /** @var string   Head section of html */
        protected $sHead = '';

        /** @var string   Area to search for css and js files to replace */
        public $sSearchArea = '';

        /** @var string   Html of full page */
        public $sBody = '';

        /** @var string   CSS files with @import statements to be combined */
        protected $aImports = array();

        /** @var array    Array of css or js urls taken from head */
        protected $aLinks = array();

        /** @var array    Array of css media types */
        public $aMedia = array();

        /** @var array    Array of ordered js files */
        protected $aOrder = array();

        /** @var integer  First index for $alinks */
        protected $iCnt = 2000;

        /** @var array    Array of arguments to be used in callback functions */
        protected $aCallbackArgs = array();

        /** @var object   Holds reference to JURI object */
        public $oUri = '';

        /** @var array    Array of options for HTML minifier */
        protected $aOptions = array();

        /** @var string   Line end character used in document */
        protected $sLnEnd = "\n";

        /** @var string   Tab character used in document */
        protected $sTab;

        /**
         * Triggered by onAfterRender event; will remove all urls of css and js
         * files in head tags (except those excluded) and aggregate in single css or
         * js file
         *
         */
        public function onAfterRender()
        {
                if ($this->params->get('debug', '0'))
                {
                        global $_PROFILER;
                        JDEBUG ? $_PROFILER->mark('beforeOnAfterRender plgSystem (JCH Optimize)') : null;
                }

                $oApplication = JFactory::getApplication();
                if ($oApplication->getName() != 'site')
                {
                        return false;
                }

                $oDocument    = JFactory::getDocument();
                $sDocType     = $oDocument->getType();
                $this->sLnEnd = $oDocument->_getLineEnd();
                $this->sTab   = $oDocument->_getTab();
                $this->oUri   = clone JURI::getInstance();

                if ($sDocType != 'html')
                {
                        return false;
                }

                $iCss        = $this->params->get('css', 1);
                $iJavaScript = $this->params->get('javascript', 1);
                $iHtmlMin    = $this->params->get('html_minify', 0);

                if (!$iCss && !$iJavaScript && !$iHtmlMin)
                {
                        return false;
                }

                $aOptions = array();
                if ($this->params->get('css_minify', 0))
                {
                        if (!class_exists('JCH_Optimize\Minify_CSS'))
                        {
                                require_once ( dirname(__FILE__) . '/libs/CSS.php' );
                        }
                        $aOptions['cssMinifier'] = array('JCH_Optimize\Minify_CSS', 'process');
                }
                if ($this->params->get('js_minify', 0))
                {
                        if (!class_exists('JCH_Optimize\JSmin'))
                        {
                                require_once ( dirname(__FILE__) . '/libs/JSMin.php' );
                        }
                        $aOptions['jsMinifier'] = array('JCH_Optimize\JSMin', 'minify');
                }

                $this->sBody = JResponse::getBody();
                $sHeadRegex  = $this->getHeadRegex();

                $sCacheGroup  = 'plg_jch_optimize';
                $this->oCache = JFactory::getCache($sCacheGroup, 'callback', 'file');
                $this->oCache->setCaching(1);
                $this->oCache->setLifetime((int) $this->params->get('lifetime', '30') * 24 * 60 * 60);

                $this->processHTML();

                if ($iJavaScript)
                {
                        $iDefer      = $this->params->get('defer_js', 0);
                        $iJsPosition = $this->params->get('bottom_js', 1);

                        $sLink      = '<script type="text/javascript" src="URL"';
                        $sLink .= $iDefer ? ($this->isXhtml() ? ' defer="defer"' : ' defer' ) : '';
                        $sLink .= $this->getAsyncAttribute();
                        $sLink .= '></script>';
                        $sNewJsLink = $iJsPosition == 1 ? $this->sTab . $sLink . $this->sLnEnd . '</body>' : $sLink;
                        $iCnt       = $this->aCallbackArgs['counter'];
                        if (!empty($this->aLinks['js']))
                        {
                                $iJsId = $this->processLink('js', $sCacheGroup, $sNewJsLink, $iCnt);
                        }
                }

                if ($iCss || $this->params->get('csg_enable', 0))
                {
                        $sNewCssLink = '</title>' . $this->sLnEnd . $this->sTab . '<link rel="stylesheet" type="text/css" ';
                        $sNewCssLink .= 'href="URL"/>';

                        if (!empty($this->aLinks['css']))
                        {
                                $iCssId = $this->processLink('css', $sCacheGroup, $sNewCssLink);
                        }
                }

                $sBody = $this->getBody();


                if ($iHtmlMin)
                {
                        if (!class_exists('JCH_Optimize\Minify_HTML'))
                        {
                                require_once ( dirname(__FILE__) . '/libs/HTML.php' );
                        }
                        $sBody = JCH_Optimize\Minify_HTML::minify($sBody, $aOptions);
                }
                JResponse::setBody($sBody);

                if ($this->params->get('debug', '0'))
                {
                        JDEBUG ? $_PROFILER->mark('afterOnAfterRender plgSystem (JCH Optimize)') : null;
                }
        }

        /**
         * Removes applicable js and css links from search area
         */
        public function processHTML()
        {
                if ($this->params->get('debug', '0'))
                {
                        global $_PROFILER;
                        JDEBUG ? $_PROFILER->mark('beforeProcessHTML plgSystem (JCH Optimize)') : null;
                }

                $params       = $this->params;
                $sHeadRegex   = $this->getHeadRegex();
                $aHeadMatches = array();
                preg_match($sHeadRegex, $this->sBody, $aHeadMatches);

                $this->sSearchArea = $this->getSearchArea($aHeadMatches[0]);

                unset($this->aLinks['js']);
                unset($this->aLinks['css']);

                $sExExtensionsRegex = '';

                if ($params->get('excludeAllExtensions', '1'))
                {
                        $sExExtensionsRegex = '|(?:\Wcomponents\W)|(?:\Wmodules\W)|(?:\Wplugins\W)';
                }

                $sExJsCompRegex  = $this->getExCompRegex($params->get('excludeJsComponents', ''));
                $sExCssCompRegex = $this->getExCompRegex($params->get('excludeCssComponents', ''));

                $this->aCallbackArgs['counter'] = 0;
                $this->excludeIf();

                if ($params->get('javascript', 1))
                {
                        $sType                                 = 'js';
                        $this->aCallbackArgs['type']           = $sType;
                        $this->aCallbackArgs['excludes']['js'] = self::getArray($params->get('excludeJs', ''));

                        $aJsRegex = $this->getJsRegex();

                        $sJsRegex          = $aJsRegex['start'] . $sExExtensionsRegex . $sExJsCompRegex . $aJsRegex['end'];
                        $this->sSearchArea = preg_replace_callback($sJsRegex,
                                                                   array(
                                __CLASS__,
                                'replaceScripts'), $this->sSearchArea);
                }

                if ($params->get('css', 1) || $params->get('csg_enable', 0))
                {
                        $sType                                  = 'css';
                        $this->aCallbackArgs['type']            = $sType;
                        $this->aCallbackArgs['excludes']['css'] = self::getArray($params->get('excludeCss', ''));

                        $aCssRegex = $this->getCssRegex();

                        $sCssRegex         = $aCssRegex['start'] . $sExExtensionsRegex . $sExCssCompRegex . $aCssRegex['end'];
                        $this->sSearchArea = preg_replace_callback($sCssRegex, array(__CLASS__, 'replaceScripts'), $this->sSearchArea);
                }

                if ($this->params->get('debug', '0'))
                {
                        JDEBUG ? $_PROFILER->mark('afterProcessHTML plgSystem (JCH Optimize)') : null;
                }
        }

        /**
         * Determine if document is of XHTML doctype
         * 
         * @return boolean
         */
        protected function isXhtml()
        {
                if (preg_match('~^<!DOCTYPE(?=[^>]+XHTML)~i', trim($this->sBody)) === 1)
                {
                        return true;
                }
                else
                {
                        return false;
                }
        }

        /**
         * Generates regex for excluding components set in plugin params
         * 
         * @param string $param
         * @return string
         */
        protected function getExCompRegex($param)
        {
                $sExCompRegex = '';
                if (isset($param) && $param)
                {
                        $aComponents = self::getArray($param);
                        foreach ($aComponents as $sComponent)
                        {
                                $sExCompRegex .= '|(?:/' . $sComponent . '/)';
                        }
                }

                return $sExCompRegex;
        }

        /**
         * Add js and css urls in conditional tags to excludes list
         *
         * @param string $sType   css or js
         */
        protected function excludeIf()
        {
                if (preg_match_all('~<\!--.*?-->~is', $this->sSearchArea, $aMatches))
                {
                        foreach ($aMatches[0] as $sMatch)
                        {
                                preg_match_all('~<(?:link|style|script)[^>]*>(?:(?:.*?)</(?:script|style)>)?~is', $sMatch, $aExcludesMatches);
                                foreach ($aExcludesMatches[0] as $sExcludesMatch)
                                {
                                        $this->aCallbackArgs['excludes']['if'][] = @$sExcludesMatch;
                                }
                        }
                }
        }

        /**
         * Use generated id to cache aggregated file
         *
         * @param string $sType           css or js
         * @param string $sCacheGroup    Name of cache group
         * @param string $sLink           Url for aggregated file
         */
        protected function processLink($sType, $sCacheGroup, $sLink, $iCnt = '')
        {

                $sId       = md5(serialize($this->aLinks[$sType]) . $this->params);
                $aArgs     = array($this->aLinks[$sType], $sType);
                $aFunction = array($this, 'getContents');

                $bCached   = $this->loadCache($aFunction, $aArgs, $sId);
                $sFileName = $this->getFilename($sId, $sCacheGroup);

                $iTime = (int) $this->params->get('lifetime', '30');

                if ($bCached)
                {
                        $sUrl     = $this->buildUrl($sFileName, $sType, $iTime, $this->isGZ());
                        $sNewLink = str_replace('URL', $sUrl, $sLink);
                        $this->replaceLink($sNewLink, $sType, $iCnt);
                }
//print_r($sFile);
                return $sId;
        }

        /**
         * Returns url of aggregated file
         *
         * @param string $sFile		Aggregated file name
         * @param string $sType		css or js
         * @param mixed $bGz		True (or 1) if gzip set and enabled
         * @param number $sTime		Expire header time
         * @return string			Url of aggregated file
         */
        protected function buildUrl($sFile, $sType, $iTime, $bGz = false)
        {
                $sPath = JURI::base(true) . '/plugins/system/jch_optimize/assets/';
                if ($this->params->get('htaccess', 0))
                {
                        $sUrl = $sPath . ($bGz ? 'gz/' : 'nz/') . $iTime . '/' . $sFile . '.' . $sType;
                }
                else
                {
                        $oUri = $this->oUri;
                        $oUri->setPath($sPath . 'jscss.php');

                        $aVar         = array();
                        $aVar['f']    = $sFile;
                        $aVar['type'] = $sType;
                        if ($bGz)
                        {
                                $aVar['gz'] = 'gz';
                        }
                        $aVar['d'] = $iTime;
                        $oUri->setQuery($aVar);

                        $sUrl = htmlentities($oUri->toString(array('path', 'query')));
                }
                return ($sUrl);
        }

        /**
         * Insert url of aggregated file in html
         *
         * @param string $sNewLink   Url of aggregated file
         */
        protected function replaceLink($sNewLink, $sType, $iCnt = '')
        {
                if ($sType == 'css')
                {
                        $this->sSearchArea = str_replace('</title>', $sNewLink, $this->sSearchArea);
                }
                if ($sType == 'js')
                {
                        switch ($this->params->get('bottom_js', 1))
                        {
                                case 0: //First found javascript tag
                                        $this->sSearchArea = preg_replace('~<JCH_SCRIPT>~', $sNewLink, $this->sSearchArea, 1);
                                        $this->sSearchArea = str_replace('<JCH_SCRIPT>', '', $this->sSearchArea);
                                        break;
                                case 2: //Last found javascript tag
                                        $iCnt--;
                                        $this->sSearchArea = preg_replace('~<JCH_SCRIPT>~', '', $this->sSearchArea, $iCnt);
                                        $this->sSearchArea = str_replace('<JCH_SCRIPT>', $sNewLink, $this->sSearchArea);
                                        break;
                                case 1: //Bottom of page
                                        $this->sSearchArea = str_replace('<JCH_SCRIPT>', '', $this->sSearchArea);
                                        $this->sSearchArea = str_replace('</body>', $sNewLink, $this->sSearchArea);
                                        $this->sBody       = str_replace('</body>', $sNewLink, $this->sBody);
                                        break;

                                default:
                                        break;
                        }
                }
        }

        /**
         * Create and cache aggregated file if it doesn't exists, file will have
         * lifetime set in global configurations.
         *
         * @param array $aFunction    Name of function used to aggregate files
         * @param array $aArgs        Arguments used by function above
         * @param string $sId         Generated id to identify cached file
         * @return boolean           True on success
         */
        protected function loadCache($aFunction, $aArgs, $sId)
        {
                unset($bCached);
                $bCached = $this->oCache->get($aFunction, $aArgs, $sId);
                if (isset($bCached))
                {
                        return true;
                }
                return false;
        }

        /**
         * Gets name of aggregated file
         *
         * @param string $sId			Id of cached file
         * @param string $sCacheGroup	Name of cache group
         * @return string				Cache file name
         */
        protected function getFilename($sId, $sCacheGroup)
        {
                $oStorage = $this->oCache->_getStorage();
                $sName1   = md5($oStorage->_application . '-' . $sId . '-' . $oStorage->_language);
                $sName    = $oStorage->_hash . '-cache-' . $sCacheGroup . '-' . $sName1;
                return $sName;
        }

        /**
         * Check if gzip is set or enabled
         *
         * @return boolean   True if gzip parameter set and server is enabled
         */
        public function isGZ()
        {
                $iIsGz = $this->params->get('gzip', 0);

                switch (TRUE)
                {
                        case (!$iIsGz):
                        case (!isset($_SERVER['HTTP_ACCEPT_ENCODING'])):
                        case (!extension_loaded('zlib') || ini_get('zlib.output_compression')):
                                return false;
                        case (false !== strpos($_SERVER['HTTP_ACCEPT_ENCODING'], 'gzip')):
                                return true;
                        default:
                                return false;
                }
        }

        /**
         * Callback function used to remove urls of css and js files in head tags
         *
         * @param array $aMatches Array of all matches
         * @return string        Returns the url if excluded, empty string otherwise
         */
        protected function replaceScripts($aMatches)
        {

                if (!isset($aMatches[2]) && !isset($aMatches[3]))
                {
                        return $aMatches[0];
                }
                $sType   = $this->aCallbackArgs['type'];
                $sEditor = JFactory::getUser()->getParam('editor');

                $sUrl         = $aMatches[2];
                $sFile        = $aMatches[1];
                $sDeclaration = isset($aMatches[3]) ? $aMatches[3] : null;
                $aExcludes    = array();
                if ($this->aCallbackArgs)
                {
                        $aExcludes = $this->aCallbackArgs['excludes'];
                }
                if ($sUrl != '')
                {
                        $sPath = self::getFilepath($sUrl);
                }

                $sMedia = '';
                if (($sType == 'css') && (preg_match('~media=(?(?=["\'])(?:["\']([^"\']+))|(\w+))~i', $aMatches[0], $aMediaTypes) > 0))
                {
                        $sMedia .= $aMediaTypes[1] ? $aMediaTypes[1] : $aMediaTypes[2];
                }

                switch (TRUE)
                {
                        case (isset($aExcludes['if']) && in_array($aMatches[0], $aExcludes['if'])):
                        case (($sFile != '') && in_array($sFile, $aExcludes[$sType])):
                        case ($sEditor == 'artofeditor' && $sFile == 'ckeditor.js'):
                        case (($sType == 'js') && ($sUrl != '') && $this->isEditorsExcluded($sUrl)):
                        case (($sUrl != '') && $this->isUrlFopenAllowed($sUrl) && !file_exists($sPath)):
                        case (isset($aMatches[3]) && (preg_match('~document\.write~i', $aMatches[3]) === 1)):
                                return $aMatches[0];

                        case ($sUrl == '' && isset($aMatches[3])):
                                $content                = str_replace(array('<!--', '-->'), '', $aMatches[3]);
                                $this->aLinks[$sType][] = array('content' => $content, 'match' => $aMatches[0]);
                                return '';

                        case ($this->aCallbackArgs['type'] == 'js'):
                                $this->aCallbackArgs['counter'] ++;
                                $this->aLinks[$sType][] = array(
                                        'url'   => $sUrl,
                                        'file'  => $sFile,
                                        'match' => $aMatches[0],
                                        'path'  => $sPath);
                                return '<JCH_SCRIPT>';

                        default:
                                $this->aLinks[$sType][] = array(
                                        'url'   => $sUrl,
                                        'file'  => $sFile,
                                        'media' => $sMedia,
                                        'match' => $aMatches[0],
                                        'path'  => $sPath);

                                return '';
                }
        }

        /**
         * Get aggregated and possibly minified content from js and css files
         *
         * @param array $aUrlArray      Array of urls of css or js files for aggregation
         * @param string $sType         css or js
         * @return string               Aggregated (and possibly minified) contents of files
         */
        public function getContents($aUrlArray, $sType)
        {

                $sContents = $this->combineFiles($aUrlArray, $sType);
                $sContents = $this->prepareContents($sContents, $sType);
                return $sContents;
        }

        /**
         * Remove placeholders from aggregated file for caching
         * 
         * @param string $sContents       Aggregated file contents
         * @param string $sType           js or css
         * @return string
         */
        public function prepareContents($sContents, $sType)
        {
                if ($sType == 'css')
                {
                        $sContents = trim($this->sortImports($sContents));

                        if ($this->params->get('csg_enable', 0))
                        {
                                $sContents = $this->getSprite($sContents);
                        }
                }

                $sContents = str_replace(
                        array(
                        'COMMENT_START',
                        'COMMENT_IMPORT_START',
                        'COMMENT_END',
                        'DELIMITER'
                        ),
                        array(
                        $this->sLnEnd . $this->sLnEnd . '/** ',
                        $this->sLnEnd . $this->sLnEnd . '/** @import url',
                        ' */' . $this->sLnEnd . $this->sLnEnd,
                        ''
                        ), $sContents);
                return $sContents;
        }

        /**
         * Aggregate contents of CSS and JS files
         *
         * @param array $aUrlArray      Array of links of files
         * @param string $sType          CSS or js
         * @return string               Aggregarted contents
         */
        public function combineFiles($aUrlArray, $sType)
        {
                if ($this->params->get('debug', '0'))
                {
                        global $_PROFILER;
                        JDEBUG ? $_PROFILER->mark('beforeCombineFiles - ' . $sType . ' plgSystem (JCH Optimize)') : null;
                }

                $sContents = '';
                $bAsync    = false;
                $sAsyncUrl = '';
                foreach ($aUrlArray as $sUrl)
                {
                        $sContent = '';
                        if (isset($sUrl['url']))
                        {
                                if ($sType == 'js' && $sAsyncUrl != '')
                                {
                                        $sContents .= ' COMMENT_START ' . $sAsyncUrl . ' COMMENT_END 
                                                loadScript("' . $sAsyncUrl . '", function(){});  DELIMITER';
                                        $sAsyncUrl = '';
                                }
                                //$sPath = $this->getFilepath($sUrl['url']);
                                $sPath = $sUrl['path'];

                                if ($sType == 'js' && $this->loadAsync($sPath))
                                {
                                        $sAsyncUrl = $sUrl['url'];
                                        $bAsync    = true;
                                        continue;
                                }
                                else
                                {
                                        $sContent .= file_get_contents($sPath);
                                }
                        }
                        else
                        {
                                if ($sType == 'js' && $sAsyncUrl != '')
                                {
                                        $sContents .= ' COMMENT_START ' . $sAsyncUrl . ' COMMENT_END 
                                                loadScript("' . $sAsyncUrl . '", function(){' . $this->sLnEnd . $sUrl['content'] . $this->sLnEnd . '});  
                                        DELIMITER';
                                        $sAsyncUrl = '';
                                }
                                else
                                {
                                        $sContent .= $sUrl['content'];
                                }
                        }

                        $sContent = trim($sContent);

                        if ($sType == 'css')
                        {
                                unset($this->aCallbackArgs['css_url']);
                                if (isset($sUrl['url']))
                                {
                                        $this->aCallbackArgs['css_url'] = $sUrl['url'];
                                }
                                $sContent = preg_replace('~@import (?:url\()?[\'"]([^\'"]+)[\'"](?:\))?~', '@import url($1)', $sContent);
                                $sContent = $this->correctUrl($sContent);
                                $sContent = $this->replaceImports($sContent);

                                if (isset($sUrl['media']) && ($sUrl['media'] != ''))
                                {
                                        $sContent = '@media ' . $sUrl['media'] . ' {' . $this->sLnEnd . $sContent . $this->sLnEnd . ' }';
                                }

                                if ($this->params->get('css_minify', 0) && preg_match_all('~[\n\r\s]~', substr($sContent, -100), $aMatches) > 5)
                                {
                                        if ($this->params->get('debug', '0'))
                                        {
                                                JDEBUG ? $_PROFILER->mark('beforeMinifyCss - "' . $sUrl['url'] . '" plgSystem (JCH Optimize)') : null;
                                        }
                                        
                                        $sContent = JCH_Optimize\Minify_CSS::process($sContent);

                                        if ($this->params->get('debug', '0'))
                                        {
                                                JDEBUG ? $_PROFILER->mark('beforeMinifyCss - "' . $sUrl['url'] . '" plgSystem (JCH Optimize)') : null;
                                        }
                                }
                        }

                        if ($sType == 'js' && $sContent != '')
                        {

                                if ($this->params->get('js_minify', 0) && preg_match_all('~[\n\r\s]~', substr($sContent, -100), $aMatches) > 5)
                                {
                                        $sCategory = isset($sUrl['url']) ? $sUrl['url'] : 'Script Declaration';
                                        $sContent  = $this->minifyJS($sContent, $sCategory);
                                }
                                if (substr($sContent, -1) != ';')
                                {
                                        $sContent = $sContent . ';';
                                }
                        }
                        $sContents .= ' COMMENT_START ' . (isset($sUrl['url']) ? $sUrl['url'] : 'script declaration') . ' COMMENT_END ' . $sContent . 'DELIMITER';
                }

                if ($bAsync)
                {
                        $sContents = $this->getLoadScript() . $sContents;
                        if ($sAsyncUrl != '')
                        {
                                $sContents .= ' COMMENT_START ' . $sAsyncUrl . ' COMMENT_END loadScript("' . $sAsyncUrl . '", function(){});  DELIMITER';
                                $sAsyncUrl = '';
                        }
                }

                if ($this->params->get('debug', '0'))
                {
                        JDEBUG ? $_PROFILER->mark('afterCombineFiles - ' . $sType . ' plgSystem (JCH Optimize)') : null;
                }
                return $sContents;
        }

        /**
         * Minify js files, will write exceptions to log
         * 
         * @param string $sContent
         * @param string $sUrl
         * @return type
         */
        public function minifyJs($sContent, $sUrl)
        {
                jimport('joomla.log.log');
                JLog::addLogger(
                        array(
                        'text_file' => 'plg_jch_optimize.errors.php'), JLog::ALL, $sUrl
                );

                try
                {
                        if ($this->params->get('debug', '0'))
                        {
                                global $_PROFILER;
                                JDEBUG ? $_PROFILER->mark('beforeMinifyJs - "' . $sUrl . '" plgSystem (JCH Optimize)') : null;
                        }

                        $sContentMin = JCH_Optimize\JSMin::minify($sContent);

                        if ($this->params->get('debug', '0'))
                        {
                                JDEBUG ? $_PROFILER->mark('afterMinifyJs - "' . $sUrl . '" plgSystem (JCH Optimize)') : null;
                        }
                }
                catch (JCH_Optimize\JSMin_UnterminatedStringException $e)
                {
                        JLog::add(JText::_('Caught JSMin_UnterminatedStringException: ' . $e->getMessage()), JLog::WARNING, $sUrl);
                        return $sContent;
                }
                catch (JCH_Optimize\JSMin_UnterminatedCommentException $e)
                {
                        JLog::add(JText::_('Caught JSMin_UnterminatedCommentException: ' . $e->getMessage()), JLog::WARNING, $sUrl);
                        return $sContent;
                }
                catch (JCH_Optimize\JSMin_UnterminatedRegExpException $e)
                {
                        JLog::add(JText::_('Caught JSMin_UnterminatedRegExpException: ' . $e->getMessage()), JLog::WARNING, $sUrl);
                        return $sContent;
                }
                catch (Exception $e)
                {
                        JLog::add(JText::_('Caught Exception: ' . $e->getMessage()), JLog::WARNING, $sUrl);
                        return $sContent;
                }
                return $sContentMin;
        }

        /**
         * Get local path of file from the url if internal
         * If external or php file, the url is returned
         *
         * @param string  $sUrl  Url of file
         * @return string       File path
         */
        public static function getFilepath($sUrl)
        {

                $oUri     = clone JUri::getInstance();
                $sUriBase = str_replace(array('~', '/administrator'), array('\~', ''), JURI::base());
                $sUriPath = str_replace(array('~', '/administrator'), array('\~', ''), JURI::base(true));

                if (self::isInternal($sUrl) && !preg_match('~\.php~i', $sUrl))
                {
                        $sUrl = preg_replace(array('~^' . $sUriBase . '~', '~^' . $sUriPath . '/~', '~\?.*?$~'), '', $sUrl);
                        $sUrl = str_replace('/', DS, $sUrl);
                        return JPATH_ROOT . DS . $sUrl;
                }
                else
                {
                        switch (TRUE)
                        {
                                case preg_match('~://~', $sUrl):
                                        return $sUrl;
                                case (substr($sUrl, 0, 2) == '//'):
                                        return $oUri->toString(array('scheme')) . str_replace('//', '', $sUrl);
                                case (substr($sUrl, 0, 1) == '/'):
                                        return $oUri->toString(array('scheme', 'host')) . $sUrl;
                                default:
                                        return $sUriBase . $sUrl;
                                        break;
                        }
                }
        }

        /**
         * Determines if file is internal
         * 
         * @param string $sUrl  Url of file
         * @return boolean
         */
        public static function isInternal($sUrl)
        {
                $oUri = clone JUri::getInstance();
                $oUrl = JUri::getInstance($sUrl);
                //trying to resolve bug in php with parse_url before 5.4.7
                if (preg_match('~^//([^/]+)(/.*)$~i', $oUrl->getPath(), $aMatches))
                {
                        if (!empty($aMatches))
                        {
                                $oUrl->setHost($aMatches[1]);
                                $oUrl->setPath($aMatches[2]);
                        }
                        $sUrl = $oUrl->toString();
                }

                return $oUri->isInternal($sUrl);
        }

        /**
         * Splits a string into an array using any regular delimiter or whitespace
         *
         * @param string  $sString   Delimited string of components
         * @return array            An array of the components
         */
        public static function getArray($sString)
        {
                return $aArray = preg_split('~[\s,;:]+~', $sString);
        }

        /**
         * Returns url of current host
         *
         * @return string    Url of current host
         */
        public function getHost()
        {
                $sWww = $this->oUri->toString(array('scheme', 'user', 'pass', 'host', 'port'));
                return $sWww;
        }

        /**
         * Converts url of background images in css files to absolute path
         * 
         * @param string $sContent
         * @return string
         */
        protected function correctUrl($sContent)
        {
                $sContent = preg_replace_callback('~url\([\'"]?([^\'"\)]+)[\'"]?\)~i', array(__CLASS__, 'correctUrlCallback'), $sContent);

                return $sContent;
        }

        /**
         * Callback function to correct urls in aggregated css files
         *
         * @param array $aMatches Array of all matches
         * @return string         Correct url of images from aggregated css file
         */
        protected function correctUrlCallback($aMatches)
        {

                $sUriBase    = JUri::base(TRUE);
                $sImageUrl   = $aMatches[1];
                $sCssFileUrl = $this->aCallbackArgs['css_url'];


                if (!preg_match('~^/|://~', $sImageUrl))
                {
                        $aCssUrlArray = explode('/', $sCssFileUrl);
                        array_pop($aCssUrlArray);
                        $sCssRootPath = implode('/', $aCssUrlArray) . '/';
                        $sImagePath   = $sCssRootPath . $sImageUrl;
                        $oUri         = JURI::getInstance($sImagePath);
                        $sUriPath     = preg_replace('~^' . $sUriBase . '/~', '', $oUri->getPath());
                        $oUri->setPath($sUriBase . '/' . $sUriPath);
                        $sImageUrl    = $oUri->toString();
                }

                if ($this->isInternal($sCssFileUrl))
                {
                        if ($this->isInternal($sImageUrl))
                        {
                                $oUri      = JURI::getInstance($sImageUrl);
                                $sImageUrl = $oUri->toString(array('path'));
                        }
                }

                return 'url(' . $sImageUrl . ')';
        }

        /**
         * Sorts @import and @charset as according to w3C <http://www.w3.org/TR/CSS2/cascade.html> Section 6.3
         *
         * @param string $sCss       Combined css
         * @return string           CSS with @import and @charset properly sorted
         */
        protected function sortImports($sCss)
        {
                $sImportRegex  = '~@import[^;]+;~i';
                $sCharsetRegex = '~@charset[^;]+;~i';

                $n                = preg_match_all($sImportRegex, $sCss, $aImportMatches);
                $aExcludesMatches = preg_match($sCharsetRegex, $sCss, $charset);

                $sImports = '';
                foreach ($aImportMatches[0] as $sImport)
                {
                        $sImports .= $sImport . $this->sLnEnd;
                }
                $sCss    = preg_replace(array($sImportRegex, $sCharsetRegex), '', $sCss);
                $charset = $aExcludesMatches <= 0 ? '' : $charset[0];
                return $sCss    = trim($charset . $this->sLnEnd . $sImports . $sCss);
        }

        /**
         * Grabs background images with no-repeat attribute from css and merge them in one file called a sprite.
         * Css is updated with sprite url and correct background positions for affected images.
         * Sprite saved in {Joomla! base}/images/jch-optimize/
         *
         * @param string $sCss       Aggregated css file before sprite generation
         * @return string           Css updated with sprite information on success. Original css on failure
         */
        public function getSprite($sCss)
        {
                if ($this->params->get('debug', '0'))
                {
                        global $_PROFILER;
                        JDEBUG ? $_PROFILER->mark('beforeGetSprite plgSystem (JCH Optimize)') : null;
                }

                if (extension_loaded('imagick') && extension_loaded('exif'))
                {
                        $sImageLibrary = 'imagick';
                }
                else
                {
                        if (!extension_loaded('gd') || !extension_loaded('exif'))
                        {
                                return $sCss;
                        }
                        $sImageLibrary = 'gd';
                }

                if (!$aMatches = $this->processCssUrls($sCss))
                {
                        return $sCss;
                }

                $sSpritePath = JPATH_ROOT . '/images/jch-optimize';

                require_once ( dirname(__FILE__) . DS . 'libs' . DS . 'css-sprite-gen.inc.php' );

                $aFormValues                    = array();
                $aFormValues['wrap-columns']    = $this->params->get('csg_wrap_images', 'off');
                $aFormValues['build-direction'] = $this->params->get('csg_direction', 'vertical');
                $aFormValues['image-output']    = $this->params->get('csg_file_output', 'PNG');
                $aFormValues['sprite-path']     = $sSpritePath;
                $oSpriteGen                     = new JCH_Optimize\CssSpriteGen($sImageLibrary, $aFormValues);

                $aSearch = $this->generateSprite($aMatches, $oSpriteGen);
                $sCss    = str_replace($aSearch['needles'], $aSearch['replacements'], $sCss);

                if ($this->params->get('debug', '0'))
                {
                        JDEBUG ? $_PROFILER->mark('afterGetSprite plgSystem (JCH Optimize)') : null;
                }

                return $sCss;
        }

        /**
         * Generates sprite image and return background positions for image replaced with sprite
         * 
         * @param array $aMatches       Array of css declarations and image url to be included in sprite
         * @param object $oSpriteGen    Object of sprite generator
         * @return array
         */
        public function generateSprite($aMatches, $oSpriteGen)
        {
                $aDeclaration = $aMatches[0];
                $aImages      = $aMatches[1];

                $oSpriteGen->CreateSprite($aImages);
                $aSpriteCss = $oSpriteGen->GetCssBackground();

                $aPatterns        = array();
                $aPatterns[0]     = $sBgImageDecRegex = '~background-image:~';
                $aPatterns[1]     = $sBgPositionRegex = '~(background:[^;]*)\b'
                        . '((?:top|bottom|left|right|center|-?[0-9]+(?:%|[c-x]{2})?)'
                        . '\s(?:top|bottom|left|right|center|-?[0-9]+(?:%|[c-x]{2})?))([^;]*;)~';
                $aPatterns[2]     = $sBgImageRegex    = '~url\((?=[^\)]+\.(?:png|gif|jpe?g))[^\)]+\)~';
                $aPatterns[3]     = $sBgPosDecRegex   = '~background-position:[^;]+;~';

                $juri_base   = JURI::base(true);
                $sBaseUrl    = ($juri_base == '/') ? $juri_base : $juri_base . '/';
                $sSpriteName = $oSpriteGen->GetSpriteFilename();

                $aSearch = array();
                for ($i = 0; $i < count($aSpriteCss); $i++)
                {
                        if (isset($aSpriteCss[$i]))
                        {
                                $aSearch['needles'][] = $aDeclaration[$i];

                                $aReplacements    = array();
                                $aReplacements[0] = 'background:';
                                $aReplacements[1] = '$1$3';
                                $aReplacements[2] = 'url(' . $sBaseUrl . 'images/jch-optimize/' . $sSpriteName . ') ' . $aSpriteCss[$i];
                                $aReplacements[3] = '';

                                $aSearch['replacements'][] = preg_replace($aPatterns, $aReplacements, $aDeclaration[$i]);
                        }
                }
                return $aSearch;
        }

        /**
         * Uses regex to find css declarations containing background images to include in sprite
         * 
         * @param string $sCss  Aggregated css file
         * @return array        Array of css declarations and image urls to replace with sprite
         */
        public function processCssUrls($sCss = '')
        {
                if (!$sCss)
                {
                        $sCss = $this->combineFiles($this->aLinks['css'], 'css');
                }

                $params       = $this->params;
                $sRegexStart  = '
                        ~(?:{
                                (?=[^{}]+url\(
                                        (?=[^)]+\.(?:png|gif|jpe?g))
                                ([^)]+)\))
                                (?=[^}]+no-repeat)
                                (?(?=[^}]*background[^;]*(?:[:\s](?:top|bottom|left|right|center|-?[0-9]+(?:%|[c-x]{2})?)){1,2}[;\s])
                                        (?=[^}]+background[^;]*(?:[:\s](?:left|top|0(?:%|[c-x]{2})?)){2}[;\s])
                                )
                                (?:(?!(?:\srepeat(?:-(?:x|y))?)
                            ';
                $sRegexMiddle = '   
                                )[^{}])+
                        })';
                $sRegexEnd    = '~sx';


                $aIncludeImages  = self::getArray($params->get('csg_include_images'));
                $aExcludeImages  = self::getArray($params->get('csg_exclude_images'));
                $sIncImagesRegex = '';
                if (!empty($aIncludeImages[0]))
                {
                        $sIncImagesRegex .= '
                                |({
                                        (?=[^{}]+url\(
                                                (?=[^)]+/';
                        $sIncImagesRegex .= '           (?:' . implode('|', $aIncludeImages) . ')';
                        $sIncImagesRegex .= '   )
                                        ([^)]+)\))
                                [^{}]+})';
                }
                $sExImagesRegex = '';
                if (!empty($aExcludeImages[0]))
                {
                        foreach ($aExcludeImages as $sExImage)
                        {
                                $sExImage = preg_quote($sExImage);
                                $sExImagesRegex .= '|(?:/' . $sExImage . ')';
                        }
                }

                $sRegex = $sRegexStart . $sExImagesRegex . $sRegexMiddle . $sIncImagesRegex . $sRegexEnd;

                $iResult = preg_match_all($sRegex, $sCss, $aMatches);
//print_r($aMatches);
                if (isset($aMatches[2]) || isset($aMatches[3]))
                {
                        $aMatches[0] = array_merge(array_filter($aMatches[0]), array_filter($aMatches[2]));
                        $aMatches[1] = array_merge(array_filter($aMatches[1]), array_filter($aMatches[3]));
                }
                return $aMatches;
        }

        /**
         * Fetches Class property containing array of matches of urls to be removed from HTML
         * 
         * @return array
         */
        public function getReplacedFiles()
        {
                return $this->aLinks;
        }

        /**
         * Regex for head search area
         * 
         * @return string
         */
        public function getHeadRegex()
        {
                return '~<head[^>]*>.*?</head>~msi';
        }

}
